package liber

import (
	"errors"
	"fmt"
	"path/filepath"
	"regexp"
	"strings"

	"github.com/clee/gobipocket"
	"github.com/meskio/epubgo"
)

type MetadataSource struct {
	Name string
	ID   string
}

type Metadata struct {
	Title       string
	Date        string
	Description string
	ISBN        []string
	Creator     []string
	Language    []string
	Publisher   []string
	Format      []string
	Keywords    []string
	Sources     []MetadataSource
}

// Sufficient returns true if the object contains enough information.
func (m *Metadata) Sufficient() bool {
	return m.Title != ""
}

// Uniques returns the list of possible unique tokens for this book.
func (m *Metadata) Uniques() []string {
	var out []string
	for _, isbn := range m.ISBN {
		out = append(out, fmt.Sprintf("isbn:%s", isbn))
	}
	// Call it 'sig', but it's just title+author(s).
	sig := fmt.Sprintf("sig:%s|%s", m.Title, strings.Join(m.Creator, "|"))
	out = append(out, sig)
	return out
}

func parseUniqueId(u string) (*Metadata, error) {
	var m Metadata
	p := strings.SplitN(u, ":", 2)
	if len(p) < 2 {
		return nil, errors.New("unparseable unique id")
	}
	switch p[0] {
	case "isbn":
		m.ISBN = []string{p[1]}
	case "sig":
		sp := strings.Split(p[1], "|")
		if len(sp) < 2 {
			return nil, errors.New("unparseable 'sig' unique id")
		}
		m.Title = sp[0]
		if sp[1] != "" {
			m.Creator = sp[1:len(sp)]
		}
	default:
		return nil, errors.New("unknown unique id type")
	}
	return &m, nil
}

func listsOverlap(a, b []string) bool {
	for _, aa := range a {
		for _, bb := range b {
			if aa == bb {
				return true
			}
		}
	}
	return false
}

// Equals returns true if we think the two Metadata objects refer to
// the same book.
func (m *Metadata) Equals(other *Metadata) bool {
	// Compare ISBNs first, authors+title next.
	return (listsOverlap(m.ISBN, other.ISBN) ||
		(m.Title == other.Title && listsOverlap(m.Creator, other.Creator)))
}

// Merge with a more authoritative metadata source. Fields that are
// set in 'other' will take precedence over the ones in the receiver.
func (m *Metadata) Merge(other *Metadata) {
	if other.Title != "" {
		m.Title = other.Title
	}
	if other.Date != "" {
		m.Date = other.Date
	}
	if other.Description != "" {
		m.Description = other.Description
	}
	if other.Creator != nil {
		m.Creator = other.Creator
	}
	if other.ISBN != nil {
		m.ISBN = other.ISBN
	}
	if other.Publisher != nil {
		m.Publisher = other.Publisher
	}
	if other.Language != nil {
		m.Language = other.Language
	}
	if other.Format != nil {
		m.Format = other.Format
	}
	if other.Sources != nil {
		m.Sources = append(m.Sources, other.Sources...)
	}
}

var authorNameRx = regexp.MustCompile(`([\p{Lu}])[\p{Ll}]+ `)

func toYear(date string) string {
	if len(date) > 4 {
		return date[:4]
	}
	return date
}

func shortAuthorName(a string) string {
	return authorNameRx.ReplaceAllString(a, "$1. ")
}

func shortAuthorString(authors []string) string {
	var shortAuthors []string
	for _, a := range authors {
		shortAuthors = append(shortAuthors, shortAuthorName(a))
	}
	return strings.Join(shortAuthors, ", ")
}

func (m *Metadata) String() string {
	author := "unknown author"
	if len(m.Creator) > 0 {
		author = shortAuthorString(m.Creator)
	}
	year := "unknown year"
	if m.Date != "" {
		year = toYear(m.Date)
	}
	return fmt.Sprintf("%s, \"%s\", %s", author, m.Title, year)
}

func parseEpub(filename string) (*Metadata, error) {
	e, err := epubgo.Open(filename)
	if err != nil {
		return nil, err
	}
	defer e.Close()

	var m Metadata
	for _, field := range e.MetadataFields() {
		attr, err := e.MetadataAttr(field)
		if err != nil {
			continue
		}
		data, err := e.Metadata(field)
		if err != nil {
			continue
		}

		switch field {
		case "title":
			m.Title = data[0]
		case "date":
			// Just pick the first date available (year only).
			m.Date = toYear(data[0])
		case "creator":
			m.Creator = data
		case "language":
			m.Language = data
		case "publisher":
			m.Publisher = data
		case "identifier":
			for i, id := range data {
				if attr[i]["scheme"] == "ISBN" {
					m.ISBN = append(m.ISBN, id)
				}
			}
		}
	}

	m.Sources = []MetadataSource{{
		Name: "epub",
		ID:   filename,
	}}

	return &m, nil
}

func parseMobi(filename string) (*Metadata, error) {
	e, err := mobipocket.Open(filename)
	if err != nil {
		return nil, err
	}

	var m Metadata
	for field, data := range e.Metadata {
		switch field {
		case "title":
			m.Title = data[0]
		case "pubdate":
			m.Date = toYear(data[0])
		case "description":
			m.Description = data[0]
		case "publisher":
			m.Publisher = data
		case "author":
			m.Creator = data
		case "isbn":
			m.ISBN = data
		}
	}

	m.Sources = []MetadataSource{{
		Name: "mobi",
		ID:   filename,
	}}

	return &m, nil
}

var alphanumRx = regexp.MustCompile(`[\pL\pN]+`)

func parseAnything(filename string) (*Metadata, error) {
	// Since there is no metadata available, try to match
	// something using just tokens from the file name.
	noext := strings.TrimSuffix(filepath.Base(filename), filepath.Ext(filename))
	return &Metadata{
		Keywords: alphanumRx.FindAllString(noext, -1),
	}, nil
}

func Parse(filename string) (*Book, string, error) {
	var m *Metadata
	var err error
	ext := strings.ToLower(filepath.Ext(filename))
	switch ext {
	case ".epub":
		m, err = parseEpub(filename)
	case ".mobi":
		m, err = parseMobi(filename)
	case ".pdf":
		m, err = parseAnything(filename)
	default:
		return nil, "", errors.New("unsupported file format")
	}
	if err != nil {
		return nil, "", err
	}

	return &Book{Metadata: m}, ext, nil
}
